/*
	String-Class - A class that could be used to simplify the handling of strings
	Copyright (C) 2003, 2004  Michael "Serpedon" Walz

	This file is part of the String-Class-Library

	String-Class-Library is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	String-Class-Library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with SerpEncrypTool; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/***************************************************************************/
/*											String Klasse										*/
/*						Copyright (C) 2004,  Michael "Serpedon" Walz					*/
/*									 Letzte nderung: 12.10.04								*/
/***************************************************************************/

/***************************************************************************/
/*										unter Mitarbeit von									*/
/*									Christopher "Esleborn" Nerz							*/
/*									 Letzte nderung: 06.01.04								*/
/***************************************************************************/

/***************************************************************************/
/* Beschreibung:																				*/
/*		String Klasse, die die Verwendung von Strings ( Zeichenketten )		*/
/*		vereinfacht.																			*/
/*		Einschrnkungen:																		*/
/*			- sehr beschrnkt UNICODE-fhig												*/
/*			- max. 2147483646 Zeichen ( 2 ^ 31 - 2 )									*/
/***************************************************************************/

// Abkrzungen
// str: String
// ICmp: insensitive Compare
// cmp: Compare
// alph: Alphabet
// lwr: Lower
// upr: Upper
// del: Delete
// ch: char
// dup: duplicate

# include "Klasse.String.hpp"
# include <stdio.h>
# include <string.h>
# include <stdarg.h>
# include <ctype.h>
# include <stdlib.h>

//# define NDEBUG
# include <assert.h>

// # define _FORCE_ANSI
// Durch _FORCE_ANSI geht die Mglichkeit verloren, das Argument IGNORE_UMLAUTS
// bei den Funktionen MakeUpper und MakeLower anzugeben.

# ifdef _FORCE_ANSI
	// Falls ANSI-Standart gewnscht ist, ersetze Funktionen
	# define strdup(x) AnsiStrDup(x)
	# define strlwr(x) StrLwr(x)
	# define strupr(x) StrUpr(x)

	// Falls ANSI-Standart gewnscht ist, deklariere neue Funktionen
	char * AnsiStrDup ( const char * fa_strSource );
# endif


// Funktionsdeklaration	
// case insensitive ( A = a;  =  )
char StrICmp ( const char * fa_cString1, const char * fa_cString2 );

// case sensitive + (  = A,  = i )
char StrCmpAlph ( const char * fa_cString1, const char * fa_cString2 ); 

//  = A = a =  
char StrICmpAlph ( const char * fa_cString1, const char * fa_cString2 );

char * StrUpr ( char * fa_cString );
char * StrLwr ( char * fa_cString ); 
char * DelUmlauts ( char * fa_cString );
char * DelUmlautsAndLwr ( char * fa_cString );
char * DelUmlautsAndUpr ( char * fa_cString );

# define MIN(a, b)	(((a) < (b)) ? (a) : (b)) 
# define MAX(a, b)	(((a) > (b)) ? (a) : (b)) 

// Standartkonstruktor: Erstellt einen leeren String (""),
// der nur die abschlieende '\0' erhlt.
// Parameter: keine
// Rckgabewerte: keine
String::String ( )
{
	ma_string = new char [ 1 ];
	if ( ma_string == NULL )
		return;
	ma_string [ 0 ] = '\0';
}

// PRIVATER Konstruktor
String::String ( const long * fp_newLength )
{
	ma_string = new char [ ( * fp_newLength ) + 1 ];
}

// Erstellt einen String mit den ersten f_count Buchstaben von fa_cString
// Parameter: const char * const fa_cString, long f_count = MAX_LONG
// fa_cString: ein char-Array
// f_count: Die Anzahl der Buchstaben, die kopiert werden sollen
//    (Oder bis zur ersten '\0'-Zeichen)
// Rckgabewerte: keine
String::String ( const char * const fa_cString, long f_count /* = MAX_LONG */ )
{
	if ( fa_cString == NULL )
	{
		ma_string = new char [ 1 ];
		if ( ma_string != NULL )
			ma_string [ 0 ] = '\0';
		return;
	}

	if ( f_count < 0 )
		f_count = 0;

	long length = strlen ( fa_cString );
	length = MIN ( length, f_count );
	ma_string = new char [ length + 1 ];

	if ( ma_string == NULL )
		return;

	strncpy ( ma_string, fa_cString, length );
	ma_string [ length ] = '\0';
}

// Erstellt einen String mit den ersten f_count Buchstaben von fa_cString
// Parameter: const WCHAR * const fa_cString, long f_count = MAX_LONG
// fa_cString: ein WCHAR-Array
// f_count: Die Anzahl der Buchstaben, die kopiert werden sollen
//    (Oder bis zur ersten '\0'-Zeichen)
// Rckgabewerte: keine
String::String ( const WCHAR * const fa_cString, long f_count /* = MAX_LONG */)
{
	if ( fa_cString == NULL )
	{
		ma_string = new char [ 1 ];
		if ( ma_string != NULL )
			ma_string [ 0 ] = '\0';
		return;
	}

	if ( f_count < 0 )
		f_count = 0;

	long length = wcslen ( fa_cString );
	length = MIN ( length, f_count );
	ma_string = new char [ length + 1 ];

	if ( ma_string == NULL )
		return;

	WideCharToMultiByte ( CP_ACP, 0, fa_cString, length, ma_string, length, NULL, NULL );
	ma_string [ length ] = '\0';
}



// Erstellt einen String, indem f_number in einen String umgewandelt wird
// Parameter: const long f_number
// f_number: die Zahl, die umgewandelt werden soll.
// Rckgabewerte: keine
String::String ( const long f_number )
{
	char a_buffer [ 32 ];
	sprintf ( a_buffer, "%d", f_number );
	ma_string = strdup ( a_buffer );
}

// Kopierkonstruktor
// Parameter: const String & f_string
// f_string: String der kopiert werden soll
// Rckgabewerte: keine
String::String ( const String & f_string )
{
	ma_string = strdup ( f_string.ma_string );
}

// Dekonstruktor
// Parameter: keine
// Rckgabewerte: keine
String::~String ( )
{
	delete [ ] ma_string;
}

// Setzt den String auf f_string
// Parameter: const String & f_string
// s: Der String der zugewiesen wird
// Rckgabewerte: *this
String & String::operator= ( const String & f_string )
{
	return ( * this ) = f_string.ma_string;
}

// Setzt den String auf fa_cString
// Parameter: const char * fa_cString
// fa_cString: das char-Array, das zugewiesen wird
// Rckgabewerte: *this
String & String::operator= ( const char * fa_cString )
{
	if ( ma_string == fa_cString )
		return ( * this );
	delete [ ] ma_string;
	ma_string = strdup ( fa_cString );
	return ( * this );
}

// Setzt den String auf fa_cString
// Parameter: const WCHAR * fa_cString
// fa_cString: das WCHAR-Array, das zugewiesen wird
// Rckgabewerte: *this
String & String::operator= ( const WCHAR * fa_cString )
{
	String charString ( fa_cString );
	char * temp = ma_string;
	ma_string = charString.ma_string;
	charString.ma_string = temp;
	return ( * this );
}


// Fgt this->ma_string und f_string->ma_string zusammen und gibt das Ergebnis zurck
// Parameter: const String & f_string
// f_string: Der String der angehngt wird.
// Rckgabewerte: Ein neuer String der aus *this + f_string besteht
/*String String::operator+ ( const String & f_string ) const
{
	return ( * this ) + f_string.ma_string;
}*/

// Fgt this->ma_string und fa_cString zusammen und gibt das Ergebnis zurck
// Parameter: const char * fa_cString
// fa_cString: das char-Array, das angehngt wird.
// Rckgabewerte: Ein neuer String der aus *this + fa_cString besteht
String String::operator+ ( const char * fa_cString ) const
{
	if ( fa_cString == NULL )
		return ( * this );

	long newLength = Length ( ) + strlen ( fa_cString ); 
	String temp ( & newLength );
	if ( temp.ma_string == NULL )
		return ( * this );

	if ( ma_string == NULL )
		temp.ma_string [ 0 ] = '\0';
	else
		strcpy ( temp.ma_string, ma_string );

	strcat ( temp.ma_string, fa_cString );
	return temp;
}

// Chris behauptet, dass das hilft, wenn man ein char * + String machen will
// Ergngzung: Chris hat recht: Man kann damit char * + String machen
String operator+ ( const String & f_string1, const String & f_string2 )
{
	return f_string1.operator+ ( f_string2.ma_string );
};


// Hngt dem String f_string an
// Parameter: const String & f_string
// f_string: der String, der angehngt wird
// Rckgabewerte: *this
String & String::operator+= ( const String & f_string )
{
	return ( * this ) += f_string.ma_string;
}

// Hngt dem String das char-Array fa_cString an
// Parameter: const char * fa_cString
// fa_cString: das char-Array, das angehngt wird
// Rckgabewerte: *this
String & String::operator+= ( const char * fa_cString )
{
	if ( fa_cString == NULL )
		return ( * this );
	char * result = new char [ Length ( ) + strlen( fa_cString ) + 1 ];
	if ( result == NULL )
		return ( * this );

	if ( ma_string == NULL )
		result [ 0 ] = '\0';
	else
		strcpy ( result, ma_string );

	strcat ( result, fa_cString );
	delete [ ] ma_string;
	ma_string = result;
	return ( * this );
}

// Vergleicht this->ma_string mit f_string->ma_string
// Parameter: const String & f_string
// Vergleichsmethode: CASE_INSENITIVE | NOTE_UMLAUTS
// f_string: der String mit dem *this verglichen werden soll
// Rckgabewerte: true, bei ==
bool String::operator== ( const String & f_string ) const
{
	return StrICmp ( ma_string, f_string.ma_string ) == 0;
}

// Vergleicht this->ma_string mit fa_cString
// Parameter: const char * fa_cString
// Vergleichsmethode: CASE_INSENITIVE | NOTE_UMLAUTS
// fa_cString: das char-Array, mit dem *this verglichen werden soll
// Rckgabewerte: true, bei ==
bool String::operator== ( const char * fa_cString ) const
{
	return StrICmp ( ma_string, fa_cString ) == 0;
}

// Vergleicht this->ma_string mit f_string->ma_string
// Parameter: const String & f_string
// Vergleichsmethode: CASE_INSENITIVE | NOTE_UMLAUTS
// f_string: der String mit dem *this verglichen werden soll
// Rckgabewerte: true, bei !=
bool String::operator!= ( const String & f_string ) const
{
	return StrICmp ( ma_string, f_string.ma_string ) != 0;
}

// Vergleicht this->ma_string mit fa_cString
// Parameter: const char * fa_cString
// Vergleichsmethode: CASE_INSENITIVE | NOTE_UMLAUTS
// fa_cString: das char-Array, mit dem *this verglichen werden soll
// Rckgabewerte: true, bei !=
bool String::operator!= ( const char * fa_cString ) const
{
	return StrICmp ( ma_string, fa_cString ) != 0;
}

// Vergleicht this->ma_string mit f_string->ma_string
// Parameter: const String & f_string
// Vergleichsmethode: CASE_INSENITIVE | NOTE_UMLAUTS
// f_string: der String mit dem *this verglichen werden soll
// Rckgabewerte: true, bei <
bool String::operator< ( const String & f_string ) const
{
	return StrICmp ( ma_string, f_string.ma_string ) < 0;
}

// Vergleicht this->ma_string mit fa_cString
// Parameter: const char * fa_cString
// Vergleichsmethode: CASE_INSENITIVE | NOTE_UMLAUTS
// fa_cString: das char-Array, mit dem *this verglichen werden soll
// Rckgabewerte: true, bei <
bool String::operator<  ( const char * fa_cString ) const
{
	return StrICmp ( ma_string, fa_cString ) < 0;
}

// Vergleicht this->ma_string mit f_string->ma_string
// Parameter: const String & f_string
// Vergleichsmethode: CASE_INSENITIVE | NOTE_UMLAUTS
// f_string: der String mit dem *this verglichen werden soll
// Rckgabewerte: true, bei <=
bool String::operator<= ( const String & f_string ) const
{
	return StrICmp ( ma_string, f_string.ma_string ) <= 0;
}

// Vergleicht this->ma_string mit fa_cString
// Parameter: const char * fa_cString
// Vergleichsmethode: CASE_INSENITIVE | NOTE_UMLAUTS
// fa_cString: das char-Array, mit dem *this verglichen werden soll
// Rckgabewerte: true, bei <=
bool String::operator<= ( const char * fa_cString ) const
{
	return StrICmp ( ma_string, fa_cString ) <= 0;
}

// Vergleicht this->ma_string mit f_string->ma_string
// Parameter: const String & f_string
// Vergleichsmethode: CASE_INSENITIVE | NOTE_UMLAUTS
// f_string: der String mit dem *this verglichen werden soll
// Rckgabewerte: true, bei >
bool String::operator> ( const String & f_string ) const
{
	return StrICmp ( ma_string, f_string.ma_string ) > 0;
}

// Vergleicht this->ma_string mit fa_cString
// Parameter: const char * fa_cString
// Vergleichsmethode: CASE_INSENITIVE | NOTE_UMLAUTS
// fa_cString: das char-Array, mit dem *this verglichen werden soll
// Rckgabewerte: true, bei >
bool String::operator> ( const char * fa_cString ) const
{
	return StrICmp ( ma_string, fa_cString ) > 0;
}

// Vergleicht this->ma_string mit f_string->ma_string
// Parameter: const String & f_string
// Vergleichsmethode: CASE_INSENITIVE | NOTE_UMLAUTS
// f_string: der String mit dem *this verglichen werden soll
// Rckgabewerte: true, bei >=
bool String::operator>= ( const String & f_string ) const
{
	return StrICmp ( ma_string, f_string.ma_string ) >= 0;
}

// Vergleicht this->ma_string mit fa_cString
// Parameter: const char * fa_cString
// Vergleichsmethode: CASE_INSENITIVE | NOTE_UMLAUTS
// fa_cString: das char-Array, mit dem *this verglichen werden soll
// Rckgabewerte: true, bei >=
bool String::operator>= ( const char * fa_cString ) const
{
	return StrICmp ( ma_string, fa_cString ) >= 0;
}

// Gibt die Lnge des Strings zurck
// Parameter: keine
// Rckgabewerte: Lnge des Strings
long String::Length ( ) const
{
	if ( ma_string == NULL )
		return 0;
	return strlen ( ma_string );
}


// Gibt nCount Zeichen zurck und beginnt bei nFirst
// Parameter: const long f_start, long f_count = MAX_LONG ) const;
// f_start: Startadresse
// f_count: Anzahl der Zeichen
// Rckgabewerte: ein String, der ein Teilstring von *this ist
String String::Mid ( long f_start, long f_count /* = MAX_LONG */ ) const
{
	if ( f_start < 0 )
		f_start = 0;
	
	if ( f_count < 0 )
		f_count = 0;

	long length = Length ( );
	if ( f_start >= length || f_count == 0 )
	{
		String emptyString;
		return emptyString;
	}

	if ( f_start > length - f_count )
		f_count = length - f_start;

	String result ( & f_count );

	if ( result.ma_string == NULL )
	{
		result.ma_string = new char [ 1 ];
		if ( result.ma_string != NULL )
			result.ma_string [ 0 ] = '\0';
		return result;
	}

	memcpy ( result.ma_string, ma_string + f_start, f_count );	
	result [ f_count ] = '\0';
	return result;
}

// Sucht fa_cString im String
// Parameter: const char * fa_cString, long f_start = 0,
//				  const uchar f_type = CASE_INSENSITIVE | NOTE_UMLAUTS
// fa_cString: das char-Array, mit dem *this verglichen werden soll
// f_start: Bezeichnet die Stelle im String, bei der angefangen wird zu suchen
// f_type:	CASE_SENSITIVE oder CASE_INSENSITIVE
//				NOTE_UMLAUTS   oder IGNORE_UMLAUTS
// Rckgabewerte: Stelle des 1. Vorkommens zurck; -1 bei nicht-Erfolg
long String::Find ( const char * fa_cString, long f_start /* = 0 */,
		/* <-- */ const uchar f_type /* = CASE_INSENSITIVE | NOTE_UMLAUTS */ ) const
{
	if ( f_start < 0 )
		f_start = 0;

	if ( f_start >= Length ( ) || fa_cString == NULL )
		return -1;
	char * tempStr1 = NULL;
	char * tempStr2 = NULL;
	if ( f_type == ( CASE_INSENSITIVE | NOTE_UMLAUTS ) )
	{
		tempStr1	= StrLwr ( strdup ( ma_string + f_start ) );
	   tempStr2 = StrLwr ( strdup ( fa_cString          ) );
	}
	else if ( f_type == ( CASE_INSENSITIVE | IGNORE_UMLAUTS ) )
	{
		tempStr1	= DelUmlautsAndLwr ( strdup ( ma_string + f_start ) );
	   tempStr2 = DelUmlautsAndLwr ( strdup ( fa_cString          ) );
	}
	else if ( f_type == ( CASE_SENSITIVE | IGNORE_UMLAUTS ) )
	{
		tempStr1	= DelUmlauts ( strdup ( ma_string + f_start ) );
	   tempStr2 = DelUmlauts ( strdup ( fa_cString          ) );
	}
	
	if ( tempStr1 == NULL || tempStr2 == NULL )
	{
		delete [] tempStr1;
		delete [] tempStr2;
		char * result = strstr( ma_string + f_start, fa_cString );
		if ( result == NULL )
			return -1;
		return result - ma_string;
	}
	char * result = strstr ( tempStr1, tempStr2 );
	delete [ ] tempStr1;
	delete [ ] tempStr2;
	if ( result == NULL )
		return -1;
	else
		return ( result - tempStr1 + f_start );
}

// Gibt einen nicht-konstanten Zeiger auf this->ma_string zurck
// Dadurch wird: string[ 10 ] = '\a'; mglich
String::operator char * ( )
{
	if ( ma_string == NULL )
	{
		ma_string = new char [ 1 ];
		if ( ma_string != NULL )
			ma_string [ 0 ] = '\0';
	}
	return ma_string;
}

// Gibt einen konstanten Zeiger auf this->ma_string zurck
String::operator const char * ( ) const
{ 
	return ma_string;
}

// Wandelt den String in Grobuchstaben um
// Parameter: const uchar f_type = NOTE_UMLAUTS
// f_type: NOTE_UMLAUTS oder IGNORE_UMLAUTS
// Rckgabewerte: *this
String & String::MakeUpper ( const uchar f_type /* = NOTE_UMLAUTS */ )
{
	if ( f_type & IGNORE_UMLAUTS )
		strupr ( ma_string );
	else
		StrUpr ( ma_string );
	return ( * this );
}

// Wandelt den String in Kleinbuchstaben um
// Parameter: const uchar f_type = NOTE_UMLAUTS
// f_type: NOTE_UMLAUTS oder IGNORE_UMLAUTS
// Rckgabewerte: *this
String & String::MakeLower ( const uchar f_type /* = NOTE_UMLAUTS */ )
{
	if ( f_type & IGNORE_UMLAUTS )
		strlwr ( ma_string );
	else
		StrLwr ( ma_string );
	return ( * this );
}

// Wandelt Sonderzeichen in normale Zeichen um.
// (->A, ->e, ->i, etc.)
// Parameter: keine
// Rckgabewerte: *this
String & String::DeleteUmlauts ( )
{
	DelUmlauts ( ma_string );
	return ( * this );
}

// Entfernt bestimmte Zeichen am Anfang des Strings
// Parameter: const char * fa_chars = "\n\r \t"
// fa_chars: ein String, der die Zeilen enthlt, die entfernt werden 
// Rckgabewerte: *this
String & String::LTrim ( const char * fa_chars /* = "\n\r \t" */ )
{
	if ( fa_chars == NULL )
		return ( * this );

	int length = Length ( );
	int from = 0;
	for ( int count = 0; count < length; count++ )
	{
		if ( ! strchr ( fa_chars, ma_string [ count ] ) )
		{
			from = count;
			break;
		}
	}

	( * this ) = this->Mid ( from );
	return ( * this );
}
	
// Entfernt bestimmte Zeichen am Ende des Strings
// Parameter: const char * fa_chars = "\n\r \t"
// fa_chars: ein String, der die Zeilen enthlt, die entfernt werden 
// Rckgabewerte: *this
String & String::RTrim ( const char * fa_chars /* = "\n\r \t" */ )
{
	if ( fa_chars == NULL )
		return ( * this );

	int length = Length ( );
	int until = MAX_LONG;
	for ( int count = length - 1; count >= 0; count-- )
	{
		if ( ! strchr ( fa_chars, ma_string [ count ] ) )
		{
			until = count + 1;
			break;
		}
	}

	( * this ) = this->Mid ( 0, until );
	return ( * this );
}

// Entfernt bestimmte Zeichen an beiden Enden des Strings
// Parameter: const char * fa_chars = "\n\r \t"
// fa_chars: ein String, der die Zeilen enthlt, die entfernt werden 
// Rckgabewerte: *this
String & String::Trim ( const char * fa_chars /* = "\n\r \t" */ )
{
	return LTrim ( fa_chars ).RTrim ( fa_chars );
}

// Liest eine Zeile aus einer Datei aus
// Der Zeilenende wird zwar gelesen, aber nicht ausgeben
// Parameter: FILE * f_file, char f_lineEnd = '\n'
// f_file: Die Datei aus der gelesen wird
// f_lineEnd: Der char, der das Zeilenende markiert.
// Rckgabewerte: *this
char String::ReadLine ( FILE * f_file, char f_lineEnd /* = '\n' */ )
{
	if ( f_file == NULL )
	{
		ma_string [ 0 ] = '\0';
		return READ_EOF;
	}

	char result = READ_SUCCESS;
	long startPos = ftell ( f_file );
	long amount = 0;
	char readChar;
	while ( true )
	{
		readChar = fgetc ( f_file );
		if ( feof ( f_file ) )
		{
			result = READ_EOF;
			break;
		}
		else if ( readChar == f_lineEnd )
			break;
		amount++;
	}
	
	fseek ( f_file, startPos, SEEK_SET );

	bool noNewMem = false;
	char * temp = NULL;
	if ( ma_string != NULL && Length ( ) >= amount )
		noNewMem = true;
	else
		temp = new char [ amount + 1 ];

	if ( temp || noNewMem )
	{
		if ( ! noNewMem ) 
		{
			delete [ ] ma_string;
			ma_string = temp;
		}
		int count = 0;
		for ( ; count < amount; count++ )
			ma_string [ count ] = fgetc ( f_file );

		fgetc ( f_file ); // Zeilenende lesen, aber nicht schreiben

		ma_string [ amount ] = '\0';
		return result;
	}
	else
	{
		if ( ma_string )
			ma_string [ 0 ] = '\0';
		return LESS_MEMORY;
	}
}

// Schreibt eine Zeile in einer Datei aus
// Parameter: FILE * f_file, char f_lineEnd = '\n'
// f_file: Die Datei in die geschrieben wird
// f_lineEnd: Der char, der das Zeilenende markiert.
void String::WriteLine ( FILE * f_file, char f_lineEnd /* = '\n'*/ ) const
{
	if ( f_file == NULL )
		return;

	fputs ( ma_string, f_file );
	fputc ( f_lineEnd, f_file );
}


		
// Liest eine Wert aus eine .ini-Datei aus
// Parameter: const char * fa_section, const char * fa_key,
//				  const char * fa_iniFile, const char * fa_default = ""
// fa_section: [section]
// fa_key: key=wert
// fa_iniFile: Dateiname der .ini-Datei
// fa_default: Rckgabestring, wenn Eintrag nicht vorhanden
// Rckgabewerte: *this
String & String::GetIniString ( const char * fa_section, const char * fa_key,
		/* <-- */					  const char * fa_iniFile, const char * fa_default /* = "" */ )

{
	DWORD bufferSize = 32;
	DWORD returnValue;
	do 
	{
		delete [ ] ma_string;
		bufferSize *= 2;
		ma_string = new char [ bufferSize + 1 ];
		returnValue = GetPrivateProfileString ( fa_section, fa_key, fa_default,
				/* <-- */								 ma_string, bufferSize, fa_iniFile );
		if ( bufferSize > MAX_LONG / 2 - 1 )
			break;
	} while ( returnValue >= bufferSize - 2 );
	return ( * this );
}

// Schreibt den String in eine .ini-Datei
// Parameter: const char * fa_section, const char * fa_key,
//				  const char * fa_iniFile
// fa_section: [section]
// fa_key: key=wert
// fa_iniFile: Dateiname der .ini-Datei
// Rckgabewerte: true, wenn erfolgreich
bool String::WriteIniString ( const char * fa_section, const char * fa_key,
		/* <-- */					const char * fa_iniFile ) const
{
	// Die :: sind wichtig, weil sonst der Compiler die String::WriteIniString-Funktion
	// nehmen will. Hier wird aber die globale Funktion bentigt.
	return ::WriteIniString ( fa_section, fa_key, fa_iniFile, ma_string );
}

// Liest eine Wert aus einer .ini-Datei aus
// Parameter: const char * fa_section, const char * fa_key,
//				  const char * fa_iniFile, const char * fa_default = "" );
// fa_section: [section]
// fa_key: key=wert
// fa_iniFile: Dateiname der .ini-Datei
// fa_default: Rckgabestring, wenn Eintrag nicht vorhanden
// Rckgabewerte: String
String GetIniString ( const char * fa_section, const char * fa_key,
		/* <-- */		 const char * fa_iniFile, const char * fa_default /* = "" */ )
{
	String result;
	result.GetIniString ( fa_section, fa_key, fa_iniFile, fa_default );
	return result;
}

// Schreibt den String in eine .ini-Datei
// Parameter: const char * fa_section, const char * fa_key,
//				  const char * fa_iniFile, const char * fa_writeString = ""
// fa_section: [section]
// fa_key: key=wert
// fa_iniFile: Dateiname der .ini-Datei
// fa_writeString: String, der geschrieben werden soll
// Rckgabewerte: true, wenn erfolgreich
bool WriteIniString ( const char * fa_section, const char * fa_key,
		/* <-- */		 const char * fa_iniFile, const char * fa_writeString /* = "" */ )
{
	return WritePrivateProfileString ( fa_section, fa_key, fa_writeString, fa_iniFile ) ? true : false;
}

// Liest eine Zahl aus einer .ini-Datei aus
// Parameter: const char * fa_section, const char * fa_key,
//				  const char * fa_iniFile, const int f_default = 0
// fa_section: [section]
// fa_key: key=wert
// fa_iniFile: Dateiname der .ini-Datei
// f_default: Rckgabewert, wenn Eintrag nicht vorhanden
// Rckgabewerte: Die gelesene Zahl
int GetIniInt ( const char * fa_section, const char * fa_key,
		/* <-- */ const char * fa_iniFile, const int f_default /* = -1 */ )
{
	return GetPrivateProfileInt ( fa_section, fa_key, f_default, fa_iniFile );
}

// Schreibt eine Zahl in eine .ini-Datei
// Parameter: const char * fa_section, const char * fa_key,
//				  const char * fa_iniFile, const uint f_number );
// fa_section: [section]
// fa_key: key=wert
// fa_iniFile: Dateiname der .ini-Datei
// f_number: die Zahl, die geschrieben werden soll
// Rckgabewerte: true, wenn erfolgreich
bool WriteIniInt ( const char * fa_section, const char * fa_key,
		/* <-- */	 const char * fa_iniFile, const int f_number )
{
	String number ( f_number );
	return WriteIniString ( fa_section, fa_key, fa_iniFile, number );
}

// Verschlsselt einen String, indem BitOr angewendet wird
// Parameter: const char * fa_bitOr
// fa_bitOr: Der String, mit dem verschlsselt wird ( Wird so oft wiederhohlt, wie ntig )
// Anmerkung: Wenn zu verschlsselnde String und den Verschlsselungsstring an einen Stelle
//				  gleich sind, erfolgt keine XOR, weil sonst '\0' das Ergebnis wre
// Rckgabewerte: *this
String & String::Secure ( const char * fa_bitOr )
{
	if ( fa_bitOr == NULL || fa_bitOr [ 0 ] == '\0' )
		return ( * this );

	long len = Length ( );
	long bitOrLen = strlen ( fa_bitOr );
	long bitOrIndex;
	for ( long count = 0; count < len; count++ )
	{
		bitOrIndex = count % bitOrLen;
		if ( ma_string [ count ] != fa_bitOr [ bitOrIndex ] )
			ma_string [ count ] ^= fa_bitOr [ bitOrIndex ];
	}
	return ( * this );
}

// Ersetzt die Escape-Zeichenfolgen
// z.B.: "\\n" durch "\n"
// Parameter: keine
// Rckgabewerte: *this
String & String::ReplaceEscChars ( )
{
	::ReplaceEscChars ( ma_string );
	return ( * this );
}

// Verschlsselt einen String, indem BitOr angewendet wird
// Parameter: const char * fa_bitOr
// fa_bitOr: Der String, mit dem verschlsselt wird ( Wird so oft wiederhohlt, wie ntig )
// Anmerkung: Wenn zu verschlsselnde String und den Verschlsselungsstring an einen Stelle
//				  gleich sind, erfolgt keine XOR, weil sonst '\0' das Ergebnis wre
// Rckgabewerte: ein neuer String der aus dem alten String ^ fa_bitOr besteht.
String String::GetSecured ( const char * fa_bitOr ) const
{
	if ( fa_bitOr == NULL || fa_bitOr [ 0 ] == '\0' )
		return ( * this );

	long len = Length ( );
	String result ( & len );
	long bitOrLen = strlen ( fa_bitOr );
	long bitOrIndex;
	for ( long count = 0; count < len; count++ )
	{
		bitOrIndex = count % bitOrLen;
		if ( ma_string [ count ] != fa_bitOr [ bitOrIndex ] )
			result.ma_string [ count ] = ma_string [ count ] ^ fa_bitOr [ bitOrIndex ];
	}
	result.ma_string [ len ] = '\0';
	return result;
}


// Vergleicht this->ma_string mit fa_cString
// Parameter: const char * fa_cString, uchar f_type = CASE_INSENSITIVE | NOTE_UMLAUTS
// fa_cString: das char-Array, mit dem *this verglichen werden soll
// f_type:	CASE_SENSITIVE oder CASE_INSENSITIVE
//					NOTE_UMLAUTS   oder IGNORE_UMLAUTS
// Rckgabewerte: <0 wenn *this < fa_cString
//                =0 wenn *this = fa_cString
//                >0 wenn *this > fa_cString
char String::Compare ( const char * fa_cString, const uchar f_type /* = CASE_INSENSITIVE | NOTE_UMLAUTS */ ) const
{
	if ( f_type == ( CASE_INSENSITIVE | NOTE_UMLAUTS ) )
		return StrICmp ( ma_string, fa_cString );
	else if ( f_type == ( CASE_INSENSITIVE | IGNORE_UMLAUTS ) )
		return StrICmpAlph ( ma_string, fa_cString );
	else if ( f_type == ( CASE_SENSITIVE | IGNORE_UMLAUTS ) )
		return StrCmpAlph ( ma_string, fa_cString );
	else
		return strcmp ( ma_string, fa_cString );
}

// Ersetzt jedes Vorkommen von f_oldChar durch f_newChar
// Parameter: const char f_oldChar, const char f_newChar
// f_oldChar: das Zeichen, das durch chNew ersetzt wird
// f_newChar: das Zeichen, mit dem chOld ersetzt wird
// Rckgabewerte: Anzahl der Ersetzungen
long String::Replace( const char f_oldChar, const char f_newChar )
{
	if ( ma_string == NULL )
		return 0;
	long result = 0;
	for ( char * pos = ma_string; ( * pos ) != '\0'; ++pos )
	{
		if ( ( * pos ) == f_oldChar )
		{
			( * pos ) = f_newChar;
			result++;
		}
	}
	return result;
}

// Ersetzt jedes Vorkommen von fa_oldString durch fa_newString
// Parameter:	const char * fa_oldString, const char * fa_newString,
//					const uchar f_type = CASE_SENSITIVE | NOTE_UMLAUTS
// fa_oldString: der String, der durch szNew ersetzt wird
// fa_newString: der String, mit dem szOld ersetzt wird
// f_type:	CASE_SENSITIVE oder CASE_INSENSITIVE
//				NOTE_UMLAUTS   oder IGNORE_UMLAUTS
// Rckgabewerte: Anzahl der Ersetzungen
long String::Replace ( const char * fa_oldString, const char * fa_newString,
		/* <-- */  const uchar f_type /* = CASE_SENSITIVE | NOTE_UMLAUTS */)
{
	if ( ma_string == NULL || fa_oldString == NULL )
		return 0;
	char zero = '\0';
	if ( fa_newString == NULL )
		fa_newString = & zero;
	String * cmpStr = NULL;
	bool cmpStrDel = false;
	
	char * replaceStr = NULL;
	bool replaceStrDel = false;
	if ( f_type == ( CASE_INSENSITIVE | NOTE_UMLAUTS ) )
	{
		cmpStr = new String ( * this );
		StrLwr ( cmpStr->ma_string );
		replaceStr = StrLwr ( strdup ( fa_oldString ) );
		cmpStrDel = true;
		replaceStrDel = true;
	}
	else if ( f_type == ( CASE_INSENSITIVE | IGNORE_UMLAUTS ) )
	{
		cmpStr = new String( ( * this ) );
		DelUmlautsAndLwr ( cmpStr->ma_string );
		replaceStr = DelUmlautsAndLwr ( strdup ( fa_oldString ) );
		cmpStrDel = true;
		replaceStrDel = true;
	}
	else if ( f_type == ( CASE_SENSITIVE | IGNORE_UMLAUTS ) )
	{
		cmpStr = new String( ( * this ) );
		DelUmlauts ( cmpStr->ma_string );
		replaceStr = DelUmlauts ( strdup ( fa_oldString ) );
		cmpStrDel = true;
		replaceStrDel = true;
	}

	if ( cmpStr == NULL || replaceStr == NULL )
	{
		if ( cmpStrDel )
			delete cmpStr;

		if ( replaceStrDel )
			delete [ ] replaceStr;

		cmpStr = this;
		replaceStr = ( char * ) ( ( void * ) fa_oldString );
		// Anderst kann der konstante Zeiger fa_oldString nicht replaceStr zugewiesen werden.
		// Durch die Variable replaceStrDel wird allerdings sichergestellt, dass er von der
		// Funktion nicht verndert wird.

		cmpStrDel = false;
		replaceStrDel = false;
	}

	String result;
	long amount = 0;
	long lengthOld = strlen ( fa_oldString );
	long pos, oldPos;
	for ( pos = -lengthOld, oldPos = -lengthOld;
	      ( pos = cmpStr->Find ( replaceStr, pos + lengthOld, CASE_SENSITIVE | NOTE_UMLAUTS ) ) != -1;
			oldPos = pos )
	{
		result += Mid ( oldPos + lengthOld, pos - oldPos - lengthOld );
		result += fa_newString;
		amount++;
	}

	if ( cmpStrDel )
		delete cmpStr;

	if ( replaceStrDel )
		delete [ ] replaceStr;

	result += Mid ( oldPos + lengthOld );
	char * temp;
	temp = ma_string;
	ma_string = result.ma_string;
	result.ma_string = temp;
	return amount;
}

// Dreht den String um (Reverse->esreveR)
// Parameter: keine
// Rckgabewerte: *this
String & String::MakeReverse ( )
{
	long length = Length ( );
	char tempCh;
	for ( long pos = 0; pos < ( ( length + 1 ) / 2 ); pos++ )
	{
		tempCh = ma_string [ pos ];
		ma_string [ pos ]	= ma_string [ length - 1 - pos ];
		ma_string [ length - 1 - pos ] = tempCh;
	}
	return ( * this );
}

// Wandelt die Zahl f_number in einen String um und hngt diesen an
// Parameter: const long f_number
// f_number: Die Zahl, die umgewandelt und angehngt wird
// Rckgabewerte: *this
String & String::AddNumber ( const long f_number )
{
	char a_buffer [ 32 ];
	sprintf ( a_buffer, "%d", f_number );
	( * this ) += a_buffer;
	return ( * this );
}

// Lscht den Inhalt des Strings (Neuer Wert: "")
// Parameter: keine
// Rckgabewerte: *this
String & String::Clear ( )
{
	delete [ ] ma_string;
	ma_string = new char [ 1 ];
	if ( ma_string != NULL )
		ma_string [ 0 ] = '\0';
	return ( * this );
}

// Formatierung wie printf unter Benutzung des String als Ausgabemedium
// Parameter: const char * fa_format, ...
// fa_format: der Formatierungsstring
// ...: Argumente, die wegen des Formatierungsstring gebraucht werden
// Rckgabewerte: *this
String & String::Format ( const char * fa_format, ... )
{
	va_list argumente;
	va_start ( argumente, fa_format );
	FormatV ( fa_format, argumente );
	va_end ( argumente );
	return ( * this );
}

// Formatierung wie printf unter Benutzung des String als Ausgabemedium
// Parameter: const char * fa_format, va_list & f_argumente
// fa_format: der Formatierungsstring
// f_argumente: Argumente, die wegen des Formatierungsstring gebraucht werden
// Rckgabewerte: *this
String & String::FormatV ( const char * fa_format, va_list & f_arguments )
{
	if ( fa_format == NULL )
		return ( * this );
	va_list argumentsBackup = f_arguments;
	
	// Berechne die maximale Lnge des Ausgabestrings
	long guessLength = 0;
	for ( const char * scan = fa_format; ( * scan ) != '\0'; scan++ )
	{
		// Normaler Text oder %..?
		if ( ( * scan ) != '%' || ( * ++scan ) == '%' )
		{
			guessLength++;
			continue;
		}
		
		// Suche nach flags ( - + 0 blank # * )
		long width = 0; // Zum Speichern der Breite, falls ein Sternchen kommt
		for ( ; ( * scan ) != '\0'; scan++ )
		{
			if ( ( * scan ) == '#' )
				guessLength += 2; // fr '0x'
			else if ( ( * scan ) == '-' || ( * scan ) == '+' || ( * scan ) == '0' || ( * scan ) == ' ' )
				; // Keine Anpassung ntig
			else if ( ( * scan ) == '*' )
				width = va_arg ( f_arguments, int );
			else
				break; // Kein flag gefunden
		}
		assert ( width >= 0 );
		
		if ( width == 0 ) // kein Sterchen fr die Lnge angegeben
		{
			width = atoi ( scan );
			for( ; ( * scan ) != '\0' && isdigit( * scan ); ++scan )
				; // Leerzeile
		}
		assert( width >= 0 );

		long precision = 0;
		if ( ( * scan ) == '.' ) // Gibt es eine Precision-Angabe?
		{
			// auf nchstes Zeichen zeigen
			++scan;
			if ( ( * scan ) == '*' )
			{
				precision = va_arg ( f_arguments, int );
				++scan;
			} else
			{
				precision = atoi ( scan );
				for( ; ( * scan ) != '\0' && isdigit( * scan ); ++scan )
					; // Leerzeile
			}
		}
		assert ( precision >= 0 );

		bool i64 = false;
		if ( strncmp ( scan, "I64", 3 ) == 0 )
		{
			scan += 3;
			i64 = true;
		} else
		{
			switch( ( * scan ) )
			{
			case 'h':
			case 'l':
			case 'F':
			case 'N':
			case 'L':
				++scan;
				// Da die Stringklasse sowieso kein Unicode untersttzt,
				// muss es hier auch nicht bercksichtig werden.
				break;
			}
		}

		long partLength = 0;
		
		// Endlich bei den eigentlichen Platzhaltern
		switch( ( * scan ) )
		{
		case 'c':
		case 'C':
			partLength = 1;
			va_arg ( f_arguments, char );
			break;
		case 's':
		case 'S':
			{
				const char * currentString = va_arg ( f_arguments, const char * );
				if ( currentString == NULL )
					partLength = 6; // "(null)"
				else
				{
					long lengthCurrentStr = strlen ( currentString );
					partLength = MAX ( 1, lengthCurrentStr );
				}
				if ( precision != 0 )
					partLength = MIN ( partLength, precision );
				partLength = MAX ( partLength, width );
			}
			break;
		case 'd':
		case 'i':
		case 'u':
		case 'x':
		case 'X':
		case 'o':
# ifndef _FORCE_ANSI
			if ( i64 )
				va_arg ( f_arguments, __int64 );
			else
# endif
				va_arg ( f_arguments, int );
			{
				long temp = MAX ( width, 1 + precision );
				partLength = MAX ( 32, temp );
			}
			break;
		case 'e':
		case 'E':
		case 'g':
		case 'G':
			va_arg ( f_arguments, double );
			{
				long temp = MAX ( width, 10 + precision );
				partLength = MAX ( 16, temp );
			}
			break;
		case 'n': // Keine Ausgabe, n ist nur eine Kontrollfunktion
			va_arg ( f_arguments, int * );
			break;
		case 'p':
			va_arg ( f_arguments, void * );
			partLength = MAX ( 12, width );
			break;
		case 'f':
			va_arg ( f_arguments, double );
			// Grenzen fr double: 1.7E +/- 308 
			// "+" 308 + "1." + precision
			// 311 + precision
			{
				long temp = 312 + MAX ( precision, 6 );
				partLength = MAX ( temp, width );
			}
			break;
		default:
			assert ( false ); // Ein unerwarteter Platzhalter
		}
		guessLength += partLength;
	}

	char * buffer = new char [ guessLength + 1 ];
	if ( buffer == NULL )
		return ( * this );
#ifndef NDEBUG
	int result = 
#endif
	vsprintf ( buffer, fa_format, argumentsBackup );
	assert ( result <= guessLength );
	va_end ( argumentsBackup );
	( * this ) = buffer;
	delete [ ] buffer;
	return ( * this );
}

// Wandelt den String in Unicode um
// Parameter: WCHAR * * fpa_unicode
// fpa_unicode: ein Zeiger auf eine WCHAR-Array, bekommt den reservierten Speicher
//					 zugewiesen ( Dieser muss mit delete [] gelscht werden )
// Rckgabewerte: ( * fpa_unicode )
// Bsp: String test1 = "test";
//		  WCHAR * test2;
//		  test1.MakeUnicode ( & test2 );
//      // Benutzen von test2
//		  delete [ ] test2;
WCHAR * String::MakeUnicode ( WCHAR * * fpa_unicode ) const
{
	if ( fpa_unicode == NULL )
		return NULL;
	else if ( ma_string == NULL )
	{
		( * fpa_unicode ) = NULL;
		return NULL;
	}

	int lenPlusOne = strlen ( ma_string ) + 1;
	( * fpa_unicode ) = new WCHAR [ lenPlusOne ];
	
	if ( ( * fpa_unicode ) == NULL )
		return NULL;
	
	MultiByteToWideChar ( CP_ACP, 0, ma_string, lenPlusOne, ( * fpa_unicode ), lenPlusOne );
	return ( * fpa_unicode );
}

char StrICmp ( const char * fa_cString1, const char * fa_cString2 )
{
	if ( fa_cString1 == NULL || fa_cString2 == NULL )
	{
		if ( fa_cString1 != NULL )
			return 1;
		else if ( fa_cString2 != NULL )
			return -1;
		return 0;
	}
	register uchar char1, char2;
	for ( ; ( * fa_cString1 ) != '\0' && ( * fa_cString2 ) != '\0'; fa_cString1++, fa_cString2++ )
	{
		char1 = ( * fa_cString1 );
		char2 = ( * fa_cString2 );
		if ( char1 == char2 )
			continue;

		if ( ( char1 >=  97 && char1 <= 122 ) || ( char1 >= 224 && char1 <= 254 && char1 != 247 ) )
			 char1 -= 32;

		if ( ( char2 >=  97 && char2 <= 122 ) || ( char2 >= 224 && char2 <= 254 && char2 != 247 ) )
			 char2 -= 32;

		if ( char1 < char2 )
			return -1;
		else if ( char1 > char2 )
			return 1;
	}

	if ( * fa_cString1 != '\0' )
		return 1;

	if ( * fa_cString2 != '\0' )
		return -1;

	return 0;
}

char StrCmpAlph ( const char * fa_cString1, const char * fa_cString2 )
{
	if ( fa_cString1 == NULL || fa_cString2 == NULL )
	{
		if ( fa_cString1 != NULL )
			return 1;
		else if ( fa_cString2 != NULL )
			return -1;
		return 0;
	}
	
	register uchar char1, char2;
	for ( ; ( * fa_cString1 ) != '\0' && ( * fa_cString2 ) != '\0'; fa_cString1++, fa_cString2++ )
	{
		char1 = ( * fa_cString1 );
		char2 = ( * fa_cString2 );
		if ( char1 == char2 )
			continue;
		
		if ( char1 >= 224 && char1 <= 230 )
			char1 = 'a';
		else if ( char1 >= 242 && char1 <= 248 && char1 != 247 )
			char1 = 'o';
		else if ( char1 >= 249 && char1 <= 252 )
			char1 = 'u';
		else if ( char1 >= 232 && char1 <= 235 )
			char1 = 'e';
		else if ( char1 >= 236 && char1 <= 239 )
			char1 = 'i';
		else if ( char1 == 241 )
			char1 = 'n';
		else if ( char1 == 253 )
			char1 = 'y';
		else if ( char1 == 231 )
			char1 = 'c';
		else if ( char1 >= 192 && char1 <= 198 )
			char1 = 'A';
		else if ( char1 >= 210 && char1 <= 216 && char1 != 215 )
			char1 = 'O';
		else if ( char1 == 223 )
			char1 = 'S';
		else if ( char1 >= 217 && char1 <= 220 )
			char1 = 'U';
		else if ( char1 >= 200 && char1 <= 203 )
			char1 = 'E';
		else if ( char1 >= 204 && char1 <= 207 )
			char1 = 'I';
		else if ( char1 == 209 )
			char1 = 'N';
		else if ( char1 == 221 )
			char1 = 'Y';
		else if ( char1 == 199 )
			char1 = 'C';

		if ( char2 >= 224 && char2 <= 230 )
			char2 = 'a';
		else if ( char2 >= 242 && char2 <= 248 && char2 != 247 )
			char2 = 'o';
		else if ( char2 >= 249 && char2 <= 252 )
			char2 = 'u';
		else if ( char2 >= 232 && char2 <= 235 )
			char2 = 'e';
		else if ( char2 >= 236 && char2 <= 239 )
			char2 = 'i';
		else if ( char2 == 241 )
			char2 = 'n';
		else if ( char2 == 253 )
			char2 = 'y';
		else if ( char2 == 231 )
			char2 = 'c';
		else if ( char2 >= 192 && char2 <= 198 )
			char2 = 'A';
		else if ( char2 >= 210 && char2 <= 216 && char2 != 215 )
			char2 = 'O';
		else if ( char2 == 223 )
			char2 = 'S';
		else if ( char2 >= 217 && char2 <= 220 )
			char2 = 'U';
		else if ( char2 >= 200 && char2 <= 203 )
			char2 = 'E';
		else if ( char2 >= 204 && char2 <= 207 )
			char2 = 'I';
		else if ( char2 == 209 )
			char2 = 'N';
		else if ( char2 == 221 )
			char2 = 'Y';
		else if ( char2 == 199 )
			char2 = 'C';

		if ( char1 < char2 )
			return -1;
		else if ( char1 > char2 )
			return 1;
	}

	if ( * fa_cString1 != '\0' )
		return 1;

	if ( * fa_cString2 != '\0' )
		return -1;

	return 0;
}


char StrICmpAlph ( const char * fa_cString1, const char * fa_cString2 )
{
	if ( fa_cString1 == NULL || fa_cString2 == NULL )
	{
		if ( fa_cString1 != NULL )
			return 1;
		else if ( fa_cString2 != NULL )
			return -1;
		return 0;
	}
	
	register uchar char1, char2;
	for ( ; ( * fa_cString1 ) != '\0' && ( * fa_cString2 ) != '\0'; fa_cString1++, fa_cString2++ )
	{
		char1 = ( * fa_cString1 );
		char2 = ( * fa_cString2 );

		if ( char1 == char2 )
			continue;
		
		if ( ( char1 >=  97 && char1 <= 122 ) || ( char1 >= 224 && char1 <= 254 && char1 != 247 ) )
			 char1 -= 32;

		if ( char1 >= 192 && char1 <= 198 )
			char1 = 'A';
		else if ( char1 >= 210 && char1 <= 216 && char1 != 215 )
			char1 = 'O';
		else if ( char1 == 223 )
			char1 = 'S';
		else if ( char1 >= 217 && char1 <= 220 )
			char1 = 'U';
		else if ( char1 >= 200 && char1 <= 203 )
			char1 = 'E';
		else if ( char1 >= 204 && char1 <= 207 )
			char1 = 'I';
		else if ( char1 == 209 )
			char1 = 'N';
		else if ( char1 == 221 )
			char1 = 'Y';
		else if ( char1 == 199 )
			char1 = 'C';
		
		if ( ( char2 >=  97 && char2 <= 122 ) || ( char2 >= 224 && char2 <= 254 && char2 != 247 ) )
			 char2 -= 32;

		if ( char2 >= 192 && char2 <= 198 )
			char2 = 'A';
		else if ( char2 >= 210 && char2 <= 216 && char2 != 215 )
			char2 = 'O';
		else if ( char2 == 223 )
			char2 = 'S';
		else if ( char2 >= 217 && char2 <= 220 )
			char2 = 'U';
		else if ( char2 >= 200 && char2 <= 203 )
			char2 = 'E';
		else if ( char2 >= 204 && char2 <= 207 )
			char2 = 'I';
		else if ( char2 == 209 )
			char2 = 'N';
		else if ( char2 == 221 )
			char2 = 'Y';
		else if ( char2 == 199 )
			char2 = 'C';
		
		if ( char1 < char2 )
			return -1;
		else if ( char1 > char2 )
			return 1;
	}
	if ( * fa_cString1 != '\0' )
		return 1;

	if ( * fa_cString2 != '\0' )
		return -1;

	return 0;
}

char * StrUpr ( char * fa_cString )
{
	if ( fa_cString == NULL )
		return NULL;

	for ( uchar * p_ch = ( uchar * ) fa_cString; ( * p_ch ) != '\0'; p_ch++ )
	{
		if ( ( ( * p_ch ) >=  97 && ( * p_ch ) <= 122 ) || ( ( * p_ch ) >= 224 && ( * p_ch ) <= 254 && ( * p_ch ) != 247 ) )
			 ( * p_ch ) -= 32;
	}

	return fa_cString;
}

char * StrLwr ( char * fa_cString )
{
	if ( fa_cString == NULL )
		return NULL;

	for ( uchar * p_ch = ( uchar * ) fa_cString; ( * p_ch ) != '\0'; p_ch++ )
	{
		if ( ( ( * p_ch ) >=  65 && ( * p_ch ) <= 90 ) || ( ( * p_ch ) >= 192 && ( * p_ch ) <= 222 && ( * p_ch ) != 215 ) )
			 ( * p_ch ) += 32;
	}

	return fa_cString;
}

char * DelUmlauts ( char * fa_cString )
{
	if ( fa_cString == NULL )
		return NULL;

	for ( uchar * p_ch = ( uchar * ) fa_cString; ( * p_ch ) != '\0'; p_ch++ )
	{
		if ( ( * p_ch ) >= 224 && ( * p_ch ) <= 230 )
			( * p_ch ) = 'a';
		else if ( ( * p_ch ) >= 242 && ( * p_ch ) <= 248 && ( * p_ch ) != 247 )
			( * p_ch ) = 'o';
		else if ( ( * p_ch ) >= 249 && ( * p_ch ) <= 252 )
			( * p_ch ) = 'u';
		else if ( ( * p_ch ) >= 232 && ( * p_ch ) <= 235 )
			( * p_ch ) = 'e';
		else if ( ( * p_ch ) >= 236 && ( * p_ch ) <= 239 )
			( * p_ch ) = 'i';
		else if ( ( * p_ch ) == 241 )
			( * p_ch ) = 'n';
		else if ( ( * p_ch ) == 253 )
			( * p_ch ) = 'y';
		else if ( ( * p_ch ) == 231 )
			( * p_ch ) = 'c';
		else if ( ( * p_ch ) >= 192 && ( * p_ch ) <= 198 )
			( * p_ch ) = 'A';
		else if ( ( * p_ch ) >= 210 && ( * p_ch ) <= 216 && ( * p_ch ) != 215 )
			( * p_ch ) = 'O';
		else if ( ( * p_ch ) == 223 )
			( * p_ch ) = 's';
		else if ( ( * p_ch ) >= 217 && ( * p_ch ) <= 220 )
			( * p_ch ) = 'U';
		else if ( ( * p_ch ) >= 200 && ( * p_ch ) <= 203 )
			( * p_ch ) = 'E';
		else if ( ( * p_ch ) >= 204 && ( * p_ch ) <= 207 )
			( * p_ch ) = 'I';
		else if ( ( * p_ch ) == 209 )
			( * p_ch ) = 'N';
		else if ( ( * p_ch ) == 221 )
			( * p_ch ) = 'Y';
		else if ( ( * p_ch ) == 199 )
			( * p_ch ) = 'C';
	}

	return fa_cString;
}

char * DelUmlautsAndLwr ( char * fa_cString )
{
	if ( fa_cString == NULL )
		return NULL;

	for ( uchar * p_ch = ( uchar * ) fa_cString; ( * p_ch ) != '\0'; p_ch++ )
	{
		if ( ( ( * p_ch ) >=  65 && ( * p_ch ) <= 90 ) || ( ( * p_ch ) >= 192 && ( * p_ch ) <= 222 && ( * p_ch ) != 215 ) )
			 ( * p_ch ) += 32;

		if ( ( * p_ch ) >= 224 && ( * p_ch ) <= 230 )
			( * p_ch ) = 'a';
		else if ( ( * p_ch ) >= 242 && ( * p_ch ) <= 248 && ( * p_ch ) != 247 )
			( * p_ch ) = 'o';
		else if ( ( * p_ch ) >= 249 && ( * p_ch ) <= 252 )
			( * p_ch ) = 'u';
		else if ( ( * p_ch ) >= 232 && ( * p_ch ) <= 235 )
			( * p_ch ) = 'e';
		else if ( ( * p_ch ) >= 236 && ( * p_ch ) <= 239 )
			( * p_ch ) = 'i';
		else if ( ( * p_ch ) == 241 )
			( * p_ch ) = 'n';
		else if ( ( * p_ch ) == 253 )
			( * p_ch ) = 'y';
		else if ( ( * p_ch ) == 231 )
			( * p_ch ) = 'c';
	}

	return fa_cString;
}

char * DelUmlautsAndUpr ( char * fa_cString )
{
	if ( fa_cString == NULL )
		return NULL;

	for ( uchar * p_ch = ( uchar * ) fa_cString; ( * p_ch ) != '\0'; p_ch++ )
	{
		if ( ( ( * p_ch ) >=  97 && ( * p_ch ) <= 122 ) || ( ( * p_ch ) >= 224 && ( * p_ch ) <= 254 && ( * p_ch ) != 247 ) )
			 ( * p_ch ) -= 32;

		if ( ( * p_ch ) >= 192 && ( * p_ch ) <= 198 )
			( * p_ch ) = 'A';
		else if ( ( * p_ch ) >= 210 && ( * p_ch ) <= 216 && ( * p_ch ) != 215 )
			( * p_ch ) = 'O';
		else if ( ( * p_ch ) == 223 )
			( * p_ch ) = 'S';
		else if ( ( * p_ch ) >= 217 && ( * p_ch ) <= 220 )
			( * p_ch ) = 'U';
		else if ( ( * p_ch ) >= 200 && ( * p_ch ) <= 203 )
			( * p_ch ) = 'E';
		else if ( ( * p_ch ) >= 204 && ( * p_ch ) <= 207 )
			( * p_ch ) = 'I';
		else if ( ( * p_ch ) == 209 )
			( * p_ch ) = 'N';
		else if ( ( * p_ch ) == 221 )
			( * p_ch ) = 'Y';
		else if ( ( * p_ch ) == 199 )
			( * p_ch ) = 'C';
	}

	return fa_cString;
}

// Falls ANSI-Standart gewnscht ist, implementiere neue Funktionen
# ifdef _FORCE_ANSI
  char * AnsiStrDup ( const char * fa_strSource )
  {
	  if ( fa_strSource == NULL )
		  return NULL;

	  long length = strlen ( fa_strSource );
	  char * result = new char [ length + 1 ];
	  if ( result != NULL )
		  memcpy ( result, fa_strSource, length + 1 );
	  return result;
  }
# endif

// Ersetzt die Escape-Zeichenfolgen
// z.B.: "\\n" durch "\n"
void ReplaceEscChars ( char * fa_string )
{
	char * readChar = fa_string;
	char * writeChar = fa_string;

	for ( ; readChar [ 0 ] != '\0'; readChar++, writeChar++ )
	{
		if ( readChar [ 0 ] == '\\' )
		{
			readChar++;
			switch ( readChar [ 0 ] )
			{
			case 'a':
				writeChar [ 0 ] = '\a';
				break;
			case 'b':
				writeChar [ 0 ] = '\b';
				break;
			case 'f':
				writeChar [ 0 ] = '\f';
				break;
			case 'n':
				writeChar [ 0 ] = '\n';
				break;
			case 'r':
				writeChar [ 0 ] = '\r';
				break;
			case 't':
				writeChar [ 0 ] = '\t';
				break;
			case 'v':
				writeChar [ 0 ] = '\t';
				break;
			case '\'':
				writeChar [ 0 ] = '\'';
				break;
			case '\"':
				writeChar [ 0 ] = '\"';
				break;
			case '\\':
				writeChar [ 0 ] = '\\';
				break;
			case '?':
				writeChar [ 0 ] = '\?';
				break;
			case '0':
			case '1':
			case '2':
			case '3':
			case '4':
			case '5':
			case '6':
			case '7':
			case '8':
			case '9':
			{
				char number [ 4 ];
				strncpy ( number, readChar, 3 );
				number [ 3 ] = '\0';

				int iNumber = atol ( number );
				if ( iNumber >= 100 )
					readChar++;

				if ( iNumber >= 10 )
					readChar++;

				int octalNumber =	  64 * ( iNumber / 100 )
					/* <-- */	   +	8 * ( iNumber / 10 - ( iNumber / 100 ) * 10 )
					/* <-- */	   +	1 * ( iNumber % 10 );

				writeChar [ 0 ] = ( char ) octalNumber;
				break;
			}
			case 'x':
			{
				char values [ 2 ];
				for ( char count = 0; count < 2; count++ )
				{
					readChar++;
					if ( readChar [ 0 ] >= '0' && readChar [ 0 ] <= '9' )
						values [ count ] = readChar [ 0 ] - '0';
					else if ( readChar [ 0 ] >= 'a' && readChar [ 0 ] <= 'f' )
						values [ count ] = readChar [ 0 ] - 'a' + 10;
					else if ( readChar [ 0 ] >= 'A' && readChar [ 0 ] <= 'F' )
						values [ count ] = readChar [ 0 ] - 'A' + 10;
					else values [ count ] = 0;
				}
				writeChar [ 0 ] = values [ 0 ] * 16 + values [ 1 ];
			}
			default:
				readChar--;
				writeChar [ 0 ] = '\\';
			}
		}
		else
			writeChar [ 0 ] = readChar [ 0 ];
	}
	writeChar [ 0 ] = '\0';
}

int FormatSearchStr( char * search )
{
    if ( search == NULL )
        return -1;
    int iRead = 0, iWrite = 0, 
        frage = 0, prozent = 0;
    bool stern = false;
    do
    {
        switch ( search [iRead] )
        {
        case '*':
            stern = true;
            break;
        case '?':
            ++frage;
            break;
        case '%':
            ++prozent;
            break;
        default:
            if (stern || frage || prozent )
            {
                for ( ;frage > 0; --frage )
                    search[ iWrite++ ] = '?';

                if ( stern )
                {
                    search[ iWrite++ ] = '*';
                    stern = false;
                    prozent = 0;
                }
                else
                {
                    for ( ; prozent > 0; --prozent )
                        search[ iWrite++ ] = '%';
                }
                ///* frage = */ prozent = stern = 0;
            }
            search[ iWrite++ ] = search [ iRead ];
        }
    }
    while (search[ iRead++ ] != '\0' );
    
    return iRead - iWrite;
}


// gibt true zurck, wenn der search-String auf den data-String zutrifft
// Folgende Wildcards existieren:
// ?: ein beliebiges Zeichen
// *: kein, ein, oder mehrere beliebige Zeichen
// %: ein oder kein beliebiges Zeichen
// Anmerkung: f_search wird evtl. optimiert
bool MatchesSearchStr ( char * search, const char * data )
{ 
    if ( search == NULL || data == NULL )
        return false;
	FormatSearchStr ( search );

    int iSearch = 0, iData = 0;
    while ( search[ iSearch ] != '\0' )
    {
        switch ( search[ iSearch ] )
        {
        case '?':
            if ( data[ iData++ ] == '\0' )
                return false;
            ++iSearch;
            break;

        case '%':
            if ( MatchesSearchStr( & search[ ++iSearch ], & data[ iData++ ] ) )
                return true;
            return MatchesSearchStr( & search[ iSearch ], & data[ iData ] );
            break;

        case '*':
            if ( search[ ++iSearch ] == '\0')
                return true;
            while ( data[ iData ] != '\0' )
            {
                if ( MatchesSearchStr( & search[ iSearch ], & data[ iData++ ] ) )
                    return true;
            }
            return false;
            break;

        default:
            if ( data[ iData ] == '\0')
                return false;
            if ( search[ iSearch++ ] != data[ iData++ ] )
                return false;
        }
    }
    return ( data [ iData ] == '\0' );
}

// wenn erfolgreich, gibt die Anzahl der Tokens zurck, sonst -1
// Parameter: const char * fa_string, const char * fa_delims, String * fa_tokens
// fa_string: der String, der in Tokens geteilt wird
// fa_delims: Array von Trennzeichen (Bsp: ";, \t" )
// fa_tokens: Array von Strings, die die ausgelesenen Tokens empfangen;
//				  muss mindesten f_amount Elemente besitzen
// f_amount: Maximale Anzahl an Tokens
int Tokenize ( String f_string, const char * fa_delims, String * fa_tokens, int f_amount )
{
	if ( fa_delims == NULL || fa_tokens == NULL )
		return -1;
	
	char * token;
	int count;
	
	for ( count = 0, token = strtok ( f_string, fa_delims );
			count < f_amount && token != NULL;
			count++, token = strtok ( NULL, fa_delims ) )
		fa_tokens [ count ] = token;

	return count;
}


// Standartkonstruktor: Erstellt einen leeren String (""),
// der nur die abschlieende '\0' erhlt.
// Parameter: keine
// Rckgabewerte: keine
UnicodeString::UnicodeString ( )
{
	ma_string = new WCHAR [ 1 ];
	if ( ma_string == NULL )
		return;
	ma_string [ 0 ] = '\0';
}

// Erstellt einen String mit den ersten f_count Buchstaben von fa_cString
// Parameter: const char * const fa_cString, long f_count = MAX_LONG
// fa_wString: ein char-Array
// f_count: Die Anzahl der Buchstaben, die kopiert werden sollen
//    (Oder bis zur ersten '\0'-Zeichen)
// Rckgabewerte: keine
UnicodeString::UnicodeString ( const char * const fa_cString, long f_count /* = MAX_LONG */ )
{
	if ( fa_cString == NULL )
	{
		ma_string = new WCHAR [ 1 ];
		if ( ma_string != NULL )
			ma_string [ 0 ] = '\0';
		return;
	}

	if ( f_count < 0 )
		f_count = 0;

	long length = strlen ( fa_cString );
	length = MIN ( length, f_count );
	ma_string = new WCHAR [ length + 1 ];

	if ( ma_string == NULL )
		return;

	MultiByteToWideChar ( CP_ACP, 0, fa_cString, length + 1, ma_string, length + 1 );

	ma_string [ length ] = '\0';

}

// Erstellt einen UnicodeString aus f_string
// Parameter: const String & f_string
// f_string: ein String
// Rckgabewerte: keine
UnicodeString::UnicodeString ( const String & f_string ) 
{
	if ( ( const char * ) f_string == NULL )
	{
		ma_string = new WCHAR [ 1 ];
		if ( ma_string != NULL )
			ma_string [ 0 ] = '\0';
		return;
	}

	long length = strlen ( ( const char * ) f_string );
	ma_string = new WCHAR [ length + 1 ];

	if ( ma_string == NULL )
		return;

	MultiByteToWideChar ( CP_ACP, 0, ( const char * ) f_string, length + 1, ma_string, length + 1 );

	ma_string [ length ] = '\0';
}


// Kopierkonstruktor
// Parameter: const UnicodeString & f_unicodeString
// f_unicodeString: String der kopiert werden soll
// Rckgabewerte: keine
UnicodeString::UnicodeString ( const UnicodeString & f_unicodeString )
{
	ma_string = wcsdup ( f_unicodeString.ma_string );
}

// Dekonstruktor
// Parameter: keine
// Rckgabewerte: keine
UnicodeString::~UnicodeString ( )
{
	delete [ ] ma_string;
}

// Setzt den String auf fa_wString
// Parameter: const WCHAR * fa_wString
// fa_cString: das WCHAR-Array, das zugewiesen wird
// Rckgabewerte: *this
UnicodeString & UnicodeString::operator= ( const WCHAR * fa_wString )
{
	if ( ma_string == fa_wString )
		return ( * this );
	delete [ ] ma_string;
	ma_string = wcsdup ( fa_wString );
	return ( * this );
}

// Setzt den String auf f_unicodeString
// Parameter: const UnicodeString & f_unicodeString
// s: Der String der zugewiesen wird
// Rckgabewerte: *this
UnicodeString & UnicodeString::operator= ( const UnicodeString & f_unicodeString )
{
	return ( * this ) = f_unicodeString.ma_string;
}


// Gibt einen nicht-konstanten Zeiger auf this->ma_string zurck
UnicodeString::operator WCHAR * ( )
{
	if ( ma_string == NULL )
	{
		ma_string = new WCHAR [ 1 ];
		if ( ma_string != NULL )
			ma_string [ 0 ] = '\0';
	}
	return ma_string;

}

// Gibt einen konstanten Zeiger auf this->ma_string zurck
UnicodeString::operator const WCHAR * ( ) const
{
	return ma_string;
}
